package local.ws.client;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.Charset;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocketFactory;

import org.apache.commons.io.output.NullOutputStream;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class WebSocketClient {
    private final static Logger log = LoggerFactory.getLogger(WebSocketClient.class);

    private final static long CLOSING_HANDSHAKE_TIMEOUT = 2000L;
    private final static Character[] PRINTABLE_NO_DIGITS_CHARS;
    private final static Charset UTF8_CHARSET = Charset.forName("UTF-8");

    private List<WebSocketMessageListener> listeners = new ArrayList<WebSocketMessageListener>();

    private Thread readerThread;

    private boolean secure;
    private String host;
    private int port;
    private String resourceName;
    private String protocol;
    private String origin;

    private boolean reconnectEnabled;
    private int reconnectAttempts = 6;
    private long reconnectInterval = 30000L;

    private Object sync = new Object();

    private boolean closed;
    private boolean connected;
    private boolean closingHandshakeStarted;

    private Socket socket;
    private OutputStream outputStream;
    private InputStream inputStream;

    private Random rnd = new Random();

    private WebSocketReader reader;

    static {
        List<Character> chars = new ArrayList<Character>();

        for (int c = 0x21; c <= 0x2f; c++) {
            chars.add((char) c);
        }

        for (int c = 0x3a; c <= 0x7e; c++) {
            chars.add((char) c);
        }

        PRINTABLE_NO_DIGITS_CHARS = chars.toArray(new Character[chars.size()]);
    };

    public WebSocketClient(String uri) throws URISyntaxException {
        this(uri, null);
    }

    public WebSocketClient(String uri, String protocol) throws URISyntaxException {
        this(new URI(uri), protocol);
    }

    public WebSocketClient(URI uri) {
        this(uri, null);
    }

    public WebSocketClient(URI uri, String protocol) {
        this.protocol = protocol;

        validateWebSocketUri(uri);

        secure = "wss".equalsIgnoreCase(uri.getScheme());
        host = uri.getHost();
        port = uri.getPort();
        if (port == -1) {
            port = secure ? 443 : 80;
        }

        resourceName = uri.getPath();
        if (resourceName == null || resourceName.equals("")) {
            resourceName = "/";
        }

        String query = uri.getQuery();
        if (query != null) {
            resourceName = resourceName + "?" + query;
        }

        origin = "http://" + host;
    }

    public static void validateWebSocketUri(URI uri) {
        if (!uri.isAbsolute()) {
            throw new IllegalArgumentException("Absolute URL required");
        }

        String scheme = uri.getScheme().toLowerCase();
        if (!scheme.equals("ws") && !scheme.equals("wss")) {
            throw new IllegalArgumentException("Unsupported scheme: " + scheme);
        }

        if (uri.getFragment() != null) {
            throw new IllegalArgumentException("Fragment is not allowed in WS URI");
        }
    }

    public void addListener(WebSocketMessageListener listener) {
        synchronized (listeners) {
            listeners.add(listener);
        }
    }

    public void removeListener(WebSocketMessageListener listener) {
        synchronized (listeners) {
            listeners.remove(listener);
        }
    }

    public List<WebSocketMessageListener> getListeners() {
        synchronized (listeners) {
            return Collections.unmodifiableList(listeners);
        }
    }

    /**
     * @return the reconnectEnabled
     */
    public boolean isReconnectEnabled() {
        return reconnectEnabled;
    }

    /**
     * @param reconnectEnabled
     *            the reconnectEnabled to set
     */
    public void setReconnectEnabled(boolean reconnectEnabled) {
        this.reconnectEnabled = reconnectEnabled;
    }

    /**
     * @return the reconnectAttempts
     */
    public int getReconnectAttempts() {
        return reconnectAttempts;
    }

    /**
     * @param reconnectAttempts
     *            the reconnectAttempts to set
     */
    public void setReconnectAttempts(int reconnectAttempts) {
        this.reconnectAttempts = reconnectAttempts;
    }

    /**
     * @return the reconnectInterval
     */
    public long getReconnectInterval() {
        return reconnectInterval;
    }

    /**
     * @param reconnectInterval
     *            the reconnectInterval to set
     */
    public void setReconnectInterval(long reconnectInterval) {
        this.reconnectInterval = reconnectInterval;
    }

    public boolean isClosed() {
        synchronized (sync) {
            return closed || closingHandshakeStarted;
        }
    }

    public boolean isConnected() {
        synchronized (sync) {
            return connected;
        }
    }

    public void connect() throws IOException {
        synchronized (sync) {
            if (isClosed()) {
                throw new WebSocketException("WebSocket is closed");
            }
            if (isConnected()) {
                throw new WebSocketException("already connected");
            }

            // 4.1. Opening handshake

            // 4.1.2 If the user agent is not configured to use a proxy, then open a TCP connection to the host given by
            // /host/ and the port given by /port/.

            int n = 0;
            do {
                try {
                    tryConnect();
                } catch (IOException ex) {
                    log.warn("Connection failed: {}", ex.getMessage());

                    n++;
                    if (!reconnectEnabled || (reconnectAttempts > 0 && n >= reconnectAttempts)) {
                        log.error("Connection failed, reaching max number of attemps #{}", n);
                        throw ex;
                    }
                    log.info("Trying to reconnect, attemp #{}", n+1);
                    try {
                        Thread.sleep(reconnectInterval);
                    } catch (InterruptedException e) {
                        throw new WebSocketException(e);
                    }
                }
            } while (!connected);
        }
    }

    void reconnect() throws IOException {
        // try to close current connection
        try {
            close();
        } catch (IOException e) {
            // ignore
        }

        // reset all closing flags
        closed = false;
        closingHandshakeStarted = false;

        // and set connected to false;
        connected = false;

        try {
            Thread.sleep(reconnectInterval);
        } catch (InterruptedException e) {
            throw new WebSocketException(e);
        }

        connect();
    }

    private void tryConnect() throws IOException {
        socket = createSocket();

        try {
            outputStream = socket.getOutputStream();
            inputStream = socket.getInputStream();

            handshake();
        } catch (IOException ex) {
            Socket s = socket;
            socket = null;
            outputStream = null;
            inputStream = null;

            try {
                s.close();
            } catch (Exception e) {
                // ignore
            }
            throw ex;
        }

        reader = new WebSocketReader(this);
        readerThread = new Thread(reader);
        readerThread.setDaemon(true);
        readerThread.start();

        connected = true;
    }

    public void send(String str) throws IOException {
        synchronized (sync) {
            if (isClosed()) {
                throw new WebSocketException("WebSocket is closed");
            }
            if (!isConnected()) {
                throw new WebSocketException("WebSocket is not connected");
            }

            int n = 0;
            boolean sent = false;
            do {
                try {
                    trySend(str);
                    sent = true;
                } catch (IOException ex) {
                    log.warn("Sending failed: {}", ex.getMessage());

                    n++;

                    if (!reconnectEnabled || (reconnectAttempts > 0 && n >= reconnectAttempts)) {
                        log.error("Sending failed, reaching max number of attemps #{}", n);
                        throw ex;
                    }

                    reconnect();

                    log.info("Trying to resend, attemp #{}", n+1);
                }
            } while (!sent);
        }
    }

    private void trySend(String str) throws IOException {
        // Once a WebSocket connection is established, but before the WebSocket closing handshake has started, the
        // user agent must use the following steps to *send /data/ using the WebSocket*:

        // 1. Send a 0x00 byte to the server.
        outputStream.write(0x00);

        // 2. Encode /data/ using UTF-8 and send the resulting byte stream to
        // the server.
        outputStream.write(str.getBytes(UTF8_CHARSET));

        // 3. Send a 0xFF byte to the server.
        outputStream.write(0xff);
        outputStream.flush();
    }

    public void close() throws IOException {
        synchronized (sync) {
            // Once a WebSocket connection is established, the user agent must use the following steps to *start the
            // WebSocket closing handshake*.
            // These steps must be run asynchronously relative to whatever algorithm invoked this one.

            // 1. If the WebSocket closing handshake has started, then abort these steps.
            if (isClosed()) {
                return;
            }
            if (isConnected()) {
                startClosingHandshake();

                // 5. Wait a user-agent-determined length of time, or until the WebSocket connection is closed.
                try {
                    readerThread.join(CLOSING_HANDSHAKE_TIMEOUT);
                } catch (InterruptedException ex) {
                    // ignore
                }
                if (readerThread.isAlive()) {
                    log.debug("WebSocketReader Thread still alive");
                    readerThread.interrupt();
                }

                reader = null;
                readerThread = null;

                // 6. If the WebSocket connection is not already closed, then close the WebSocket connection. (If this
                // happens, then the closing handshake doesn't finish.)
                log.debug("Closing the WebSocket connection");
                socket.close();
            }

            // NOTE: The closing handshake finishes once the server returns the 0xFF packet, as described above.
            closed = true;
        }
    }

    /**
     * Starts The WebSocket closing handshake if not already started
     *
     * @return true if handshake started in this method invocation, false if closing handshake has already started.
     * @throws IOException
     */
    boolean startClosingHandshake() throws IOException {
        synchronized (sync) {
            if (closingHandshakeStarted) {
                log.debug("The WebSocket closing handshake has already started");
                return false;
            }
            // 2. Send a 0xFF byte to the server.
            outputStream.write(0xFF);
            // 3. Send a 0x00 byte to the server.
            outputStream.write(0x00);
            outputStream.flush();
            // 4. *The WebSocket closing handshake has started*.
            closingHandshakeStarted = true;
            log.debug("The WebSocket closing handshake has started");

            return true;
        }
    }

    private void handshake() throws IOException {
        OutputStreamWriter writer = new OutputStreamWriter(outputStream, Charset.forName("UTF8"));

        // 4.1.5. Send the UTF-8 string "GET" followed by a UTF-8-encoded U+0020 SPACE character to the remote side
        // (the server).
        //
        // Send the /resource name/ value, encoded as UTF-8.
        //
        // Send another UTF-8-encoded U+0020 SPACE character, followed by the UTF-8 string "HTTP/1.1", followed by a
        // UTF-8-encoded U+000D CARRIAGE RETURN U+000A LINE FEED character pair (CRLF).
        writer.write("GET ");
        writer.write(resourceName);
        writer.write(" HTTP/1.1\r\n");

        // 4.1.6. Let /fields/ be an empty list of strings.
        List<String> fields = new ArrayList<String>();

        // 4.1.7. Add the string "Upgrade: WebSocket" to /fields/.
        fields.add("Upgrade: WebSocket\r\n");

        // 4.1.8. Add the string "Connection: Upgrade" to /fields/.
        fields.add("Connection: Upgrade\r\n");

        // 4.1.9. Let /hostport/ be an empty string.
        String hostport = "";

        // 4.1.10. Append the /host/ value, converted to ASCII lowercase, to
        hostport += host;

        // 4.1.11. If /secure/ is false, and /port/ is not 80, or if /secure/ is true, and /port/ is not 443, then
        // append a U+003A COLON character (:) followed by the value of /port/, expressed as a base-ten integer, to
        // /hostport/.
        if ((!secure && port != 80) || (secure && port != 443)) {
            hostport += ":" + port;
        }

        // 4.1.12. Add the string consisting of the concatenation of the string
        // "Host:", a U+0020 SPACE character, and /hostport/, to /fields/.
        fields.add("Host: " + hostport + "\r\n");

        // 4.1.13. Add the string consisting of the concatenation of the string
        // "Origin:", a U+0020 SPACE character, and the /origin/ value, to /fields/.
        fields.add("Origin: " + origin + "\r\n");

        // 4.1.14. If there is no /protocol/, then skip this step.

        // Otherwise, add the string consisting of the concatenation of the string "Sec-WebSocket-Protocol:", a
        // U+0020 SPACE character, and the /protocol/ value, to /fields/.
        if (!StringUtils.isBlank(protocol)) {
            fields.add("Sec-WebSocket-Protocol: " + protocol + "\r\n");
        }

        SecWebSocketKey key1 = generateSecWebSocketKey();
        SecWebSocketKey key2 = generateSecWebSocketKey();

        // 4.1.23. Add the string consisting of the concatenation of the string "Sec-WebSocket-Key1:", a U+0020 SPACE
        // character, and the /key_1/ value, to /fields/.

        // Add the string consisting of the concatenation of the string "Sec-WebSocket-Key2:", a U+0020 SPACE
        // character, and the /key_2/ value, to /fields/.

        fields.add("Sec-WebSocket-Key1: " + key1.getKey() + "\r\n");
        fields.add("Sec-WebSocket-Key2: " + key2.getKey() + "\r\n");

        // 24. For each string in /fields/, in a random order: send the string, encoded as UTF-8, followed by a
        // UTF-8-encoded U+000D CARRIAGE RETURN U+000A LINE FEED character pair (CRLF). It is important
        // that the fields be output in a random order so that servers not depend on the particular order used by any
        // particular client.

        while (!fields.isEmpty()) {
            String field = fields.remove(rnd.nextInt(fields.size()));
            writer.write(field);
        }

        // 4.1.25. Send a UTF-8-encoded U+000D CARRIAGE RETURN U+000A LINE FEED character pair (CRLF).
        writer.write("\r\n");

        writer.flush();

        // 26. Let /key3/ be a string consisting of eight random bytes (or equivalently, a random 64 bit integer
        // encoded in big-endian order).
        byte[] key3 = new byte[8];
        rnd.nextBytes(key3);
        outputStream.write(key3);

        outputStream.flush();

        String code = readStatusCode(inputStream);

        // 4.1.31. If /code/, interpreted as UTF-8, is "101", then move to the next step.

        // If /code/, interpreted as UTF-8, is "407", then either close the connection and jump back to step 2,
        // providing appropriate authentication information, or fail the WebSocket connection. 407 is the code used
        // by HTTP meaning "Proxy Authentication Required". User agents that support proxy authentication must
        // interpret the response as defined by HTTP (e.g. to find and interpret the |Proxy-Authenticate| header).

        // Otherwise, fail the WebSocket connection and abort these steps.

        if (!"101".equals(code)) {
        	int c ;
        	while( (c = inputStream.read()) != -1){
        	System.out.print((char) c);
        	}
        	throw new WebSocketException("Invalid handshake response - HTTP response code: " + code );
        }

        Map<String, List<String>> fieldMap = readFields(inputStream);

        // 4.1.41. If there is not exactly one entry in the /fields/ list whose name is "upgrade", or if there is not
        // exactly one entry in the /fields/ list whose name is "connection", or if there is not exactly one entry in
        // the /fields/ list whose name is "sec-websocket-origin", or if there is not exactly one entry in the
        // /fields/ list whose name is "sec-websocket-location", or if the /protocol/ was specified but there is not
        // exactly one entry in the /fields/ list whose name is "sec-websocket-protocol", or if there are any entries
        // in the /fields/ list whose names are the empty string, then fail the WebSocket connection and abort these
        // steps. Otherwise, handle each entry in the /fields/ list as follows:

        String value;

        value = getFieldValue(fieldMap, "upgrade");
        if (!"WebSocket".equals(value)) {
            throw new WebSocketException("Expected WebSocket as Updgrade field value!");
        }
        value = getFieldValue(fieldMap, "connection");
        if (!"upgrade".equalsIgnoreCase(value)) {
            throw new WebSocketException("Expected Upgrade as Connection field value!");
        }
        value = getFieldValue(fieldMap, "sec-websocket-origin");
        if (!origin.equalsIgnoreCase(value)) {
            throw new WebSocketException("Invalid Sec-Websocket-Origin field value!");
        }
        value = getFieldValue(fieldMap, "sec-websocket-location");
        if (!constructWebSocketURL().equals(value)) {
            throw new WebSocketException("Invalid Sec-Websocket-Location field value!");
        }

        if (!StringUtils.isBlank(protocol)) {
            value = getFieldValue(fieldMap, "sec-websocket-protocol");
            if (!protocol.equals(value)) {
                throw new WebSocketException("Invalid Sec-Websocket-Protocil field value!");
            }
        }
        if (fieldMap.containsKey("")) {
            throw new WebSocketException("Empty field name found in field list!");
        }

        // 4.1.42. Let /challenge/ be the concatenation of /number_1/, expressed as a big-endian 32 bit integer,
        // /number_2/, expressed as a big-endian 32 bit integer, and the eight bytes of /key_3/ in the order they
        // were sent on the wire.
        MessageDigest digest;
        try {
            digest = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException ex) {
            throw new WebSocketException(ex);
        }
        DataOutputStream dos = new DataOutputStream(new DigestOutputStream(new NullOutputStream(), digest));
        dos.writeInt(key1.getNumber());
        dos.writeInt(key2.getNumber());
        dos.write(key3);
        dos.flush();

        // 4.1.43. Let /expected/ be the MD5 fingerprint of /challenge/ as a big-endian 128 bit string. [RFC1321]
        byte[] expected = digest.digest();
        byte[] reply = new byte[16];

        // 4.1.44. Read sixteen bytes from the server. Let /reply/ be those bytes.
        // If the connection closes before these bytes are received, then fail the WebSocket connection and abort
        // these steps.
        if (inputStream.read(reply) != reply.length) {
            throw new WebSocketException("Unexpected end of stream!");
        }

        // 4.1.45. If /reply/ does not exactly equal /expected/, then fail the WebSocket connection and abort these
        // steps.
        if (!Arrays.equals(expected, reply)) {
            throw new WebSocketException("Challenge/Response failed!");
        }

        // 4.1.46. The *WebSocket connection is established*.
        connected = true;
    }

    InputStream getInputStream() throws IOException {
        return inputStream;
    }

    private String getFieldValue(Map<String, List<String>> fieldMap, String fieldName) throws IOException {
        if (fieldMap.containsKey(fieldName)) {
            List<String> list = fieldMap.get(fieldName);
            if (list.size() == 1) {
                return list.get(0);
            }
        }
        throw new WebSocketException("Expected exacly one field '" + fieldName + "' in field list!");
    }

    private String readStatusCode(InputStream is) throws IOException {
        StringBuilder field = new StringBuilder();
        // 28. Read bytes from the server until either the connection closes, or a 0x0A byte is read. Let /field/ be
        // these bytes, including the 0x0A byte.
        int c;
        while ((c = is.read()) != -1) {
            field.append((char) c);
            if (c == '\n')
                break;
        }

        // If /field/ is not at least seven bytes long, or if the last two bytes aren't 0x0D and 0x0A respectively,
        // or if it does not contain at least two 0x20 bytes, then fail the WebSocket connection and abort these
        // steps.
        if (field.length() < 7 || !field.toString().endsWith("\r\n")) {
            throw new WebSocketException("Invalid Status Line:" + field);
        }

        // 4.1.29. Let /code/ be the substring of /field/ that starts from the byte after the first 0x20 byte, and
        // ends with the byte before the second 0x20 byte.
        Pattern p = Pattern.compile("^[^ ]* ([0-9]{3}) ");
        Matcher m = p.matcher(field);

        // 4.1.30. If /code/ is not three bytes long, or if any of the bytes in /code/ are not in the range 0x30 to
        // 0x39, then fail the WebSocket connection and abort these steps.
        if (!m.find()) {
            throw new WebSocketException("No Valid Status Code found: " + field);
        }

        String code = m.group(1);
        return code;
    }

    private Map<String, List<String>> readFields(InputStream is) throws IOException {
        // 4.1 32. Let /fields/ be a list of name-value pairs, initially empty.
        Map<String, List<String>> fields = new LinkedHashMap<String, List<String>>();

        while (true) {
            // 4.1 34. Read /name/
            String name = readFieldName(is);
            if (name.isEmpty())
                break;
            // 4.1 36. Read /value/
            String value = readFieldValue(is);

            // 4.1.38. Append an entry to the /fields/ list that has the name given by the string obtained by
            // interpreting the /name/ byte array as a UTF-8 byte stream and the value given by the string obtained
            // by interpreting the /value/ byte array as a UTF-8 byte stream.
            List<String> list = fields.get(name);
            if (list == null) {
                list = new ArrayList<String>();
                fields.put(name, list);
            }

            list.add(value);
            // 4.1 39. return to the "Field" step above
        }

        // 40. _Fields processing_: Read a byte from the server.

        // If the connection closes before this byte is received, or if the byte is not a 0x0A byte (ASCII LF), then
        // fail the WebSocket connection and abort these steps.

        // NOTE: This skips past the LF byte of the CRLF after the blank line after the fields.
        skipByte(is, 0x0A);

        return fields;
    }

    private String readFieldName(InputStream is) throws IOException {
        // This reads a field name, terminated by a colon, converting upper-case ASCII letters to lowercase, and
        // aborting if a stray CR or LF is found.

        // 4.1.33. Let /name/ be empty byte array.
        ByteArrayOutputStream name = new ByteArrayOutputStream();

        // 4.1.34. Read a byte from the server.
        // If the connection closes before this byte is received, then fail the WebSocket connection and abort these
        // steps.

        // Otherwise, handle the byte as described in the appropriate entry below:

        int b = -1;
        while ((b = is.read()) != -1) {
            if (b == 0x0D) {
                // -> If the byte is 0x0D (ASCII CR)
                // If the /name/ byte array is empty, then jump to the fields processing step. Otherwise, fail the
                // WebSocket connection and abort these steps.
                if (name.size() != 0) {
                    throw new WebSocketException("Unexpected CR byte in field name!");
                }
                break;
            } else if (b == 0x0A) {
                // -> If the byte is 0x0A (ASCII LF)
                // Fail the WebSocket connection and abort these steps.
                throw new WebSocketException("Unexpected LF byte in field name!");
            } else if (b == 0x3A) {
                // -> If the byte is 0x3A (ASCII :)
                // Move on to the next step.
                break;
            } else if (b >= 0x41 && b <= 0x5A) {
                // -> If the byte is in the range 0x41 to 0x5A (ASCII A-Z)
                // Append a byte whose value is the byte's value plus 0x20 to the /name/ byte array and redo this
                // step for the next byte.
                name.write(b + 0x20);
            } else {
                // -> Otherwise
                // Append the byte to the /name/ byte array and redo this step for the next byte.
                name.write(b);
            }
        }

        if (b == -1) {
            throw new WebSocketException("Unexpected end of stream!");
        }

        return new String(name.toByteArray(), UTF8_CHARSET);
    }

    private String readFieldValue(InputStream is) throws IOException {
        // This reads a field value, terminated by a CRLF, skipping past a single space after the colon if there is
        // one.

        // 4.1 33. Let /value/ be empty byte arrays
        ByteArrayOutputStream value = new ByteArrayOutputStream();

        // 4.1.35. Let /count/ equal 0.
        int count = 0;

        // NOTE: This is used in the next step to skip past a space character after the colon, if necessary.

        // 4.1.36. Read a byte from the server and increment /count/ by 1.

        // If the connection closes before this byte is received, then fail the WebSocket connection and abort these
        // steps.
        int b = -1;
        while ((b = is.read()) != -1) {
            count = +1;
            // Otherwise, handle the byte as described in the appropriate entry below:

            if (b == 0x20 && count == 1) {
                // -> If the byte is 0x20 (ASCII space) and /count/ equals 1
                // Ignore the byte and redo this step for the next byte.
            } else if (b == 0x0D) {
                // -> If the byte is 0x0D (ASCII CR)
                // Move on to the next step.
                break;
            } else if (b == 0x0A) {
                // -> If the byte is 0x0A (ASCII LF)
                // Fail the WebSocket connection and abort these steps.
                throw new WebSocketException("Unexpected LF character in field value!");
            } else {
                // -> Otherwise
                // Append the byte to the /value/ byte array and redo this step for the next byte.
                value.write(b);
            }
        }

        if (b == -1) {
            throw new WebSocketException("Unexpected end of stream!");
        }

        // 4.1.37. Read a byte from the server.
        // If the connection closes before this byte is received, or if the byte is not a 0x0A byte (ASCII LF), then
        // fail the WebSocket connection and abort these steps.

        // NOTE: This skips past the LF byte of the CRLF after the field.
        skipByte(is, 0x0A);

        return new String(value.toByteArray(), UTF8_CHARSET);
    }

    private void skipByte(InputStream is, int skip) throws IOException {
        int b = is.read();
        if (b == -1) {
            throw new WebSocketException("Unexpected end of stream!");
        } else if (b != skip) {
            throw new WebSocketException(String.format("Expected %X byte!", skip));
        }
    }

    private Socket createSocket() throws java.io.IOException {
        Socket socket;
        if (secure) {
            SocketFactory factory = SSLSocketFactory.getDefault();
            socket = factory.createSocket(host, port);
        } else {
            socket = new Socket(host, port);
        }
        socket.setKeepAlive(true);

        // socket.setSoTimeout(10000);
        return socket;
    }

    private class SecWebSocketKey {
        private int number;
        private String key;

        public SecWebSocketKey(int number, String key) {
            this.number = number;
            this.key = key;
        }

        /**
         * @return the number
         */
        public int getNumber() {
            return number;
        }

        /**
         * @return the key
         */
        public String getKey() {
            return key;
        }
    }

    private SecWebSocketKey generateSecWebSocketKey() {
        // 4.1 16. Let /spacesn/ be a random integer from 1 to 12 inclusive.
        int spaces = 1 + rnd.nextInt(12);

        // 4.1 17. Let /max_n/ be the largest integer not greater than 4,294,967,295 divided by /spaces_n/.
        int max = Integer.MAX_VALUE / spaces;

        // 4.1 18. Let /number_n/ be a random integer from 0 to /max_n/ inclusive.
        // JL : simplification, as nextInt(max) returns from 0 inclusive to max exclusive, but it shouldn't affect the
        // algorithm, we just missing one value of Integer.MAX_VALUE values
        int number = rnd.nextInt(max);

        // 4.1 19. Let /product_n/ be the result of multiplying /number_n/ and /spaces_n/ together.
        int product = number * spaces;

        // 4.1 20. Let /key_n/ be a string consisting of /product_n/, expressed in base ten using the numerals in the
        // range U+0030 DIGIT ZERO (0) to U+0039 DIGIT NINE (9).
        StringBuilder key = new StringBuilder(String.valueOf(product));

        // 4.1.21. Insert between one and twelve random characters from the ranges U+0021 to U+002F and U+003A to U+007E
        // into /key_n/ at random positions.

        // NOTE: This corresponds to random printable ASCII characters other than the digits and the U+0020 SPACE
        // character.
        int n = 1 + rnd.nextInt(12);
        for (int i = 0; i < n; i++) {
            int pos = rnd.nextInt(key.length());
            char ch = PRINTABLE_NO_DIGITS_CHARS[rnd.nextInt(PRINTABLE_NO_DIGITS_CHARS.length)];
            key.insert(pos, ch);
        }

        // 4.1 22. Insert /spaces_n/ U+0020 SPACE characters into /key_n/ at random positions other than start or end of
        // the string.
        for (int i = 0; i < spaces; i++) {
            int pos = 1 + rnd.nextInt(key.length() - 1);
            key.insert(pos, ' ');
        }

        return new SecWebSocketKey(number, key.toString());
    }

    private String constructWebSocketURL() {
        // 3.2. Constructing WebSocket URLs

        // The steps to *construct a WebSocket URL* from a /host/, a /port/, a /resource name/, and a /secure/ flag,
        // are as follows:

        // 3.2.1. Let /url/ be the empty string.
        String url = "";

        // 3.2.2. If the /secure/ flag is false, then append the string "ws://" to /url/. Otherwise, append the
        // string "wss://" to /url/.
        url += secure ? "wss://" : "ws://";

        // 3.2.3. Append /host/ to /url/.
        url += host;

        // 3.2.4. If the /secure/ flag is false and port is not 80, or if the /secure/ flag is true and port is not
        // 443, then append the string ":" followed by /port/ to /url/.
        if ((!secure && port != 80) || (secure && port != 443)) {
            url += ":" + port;
        }

        // 3.2.5. Append /resource name/ to /url/.
        url += resourceName;

        // 3.2.6. Return /url/.
        return url;
    }

    void onMessage(String data) {
        synchronized (listeners) {
            for (WebSocketMessageListener listener : listeners) {
                try {
                    listener.onMessage(data);
                } catch (Exception e) {
                    // ignore
                }
            }
        }
    }

    void onError(Exception ex) {
        synchronized (listeners) {
            for (WebSocketMessageListener listener : listeners) {
                try {
                    listener.onError(ex);
                } catch (Exception e) {
                    // ignore
                }
            }
        }
    }
}
