/**
 *
 */
package local.ws.client;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.concurrent.atomic.AtomicBoolean;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class WebSocketReader implements Runnable {
    private final static Logger log = LoggerFactory.getLogger(WebSocketReader.class);
    private final static Charset UTF8_CHARSET = Charset.forName("UTF-8");

    private WebSocketClient ws;
    private BufferedInputStream inputStream;
    private AtomicBoolean active = new AtomicBoolean(true);

    public WebSocketReader(WebSocketClient ws) throws IOException {
        this.ws = ws;
        inputStream = new BufferedInputStream(ws.getInputStream());
    }

    public void stop() {
        active.set(false);
    }

    //@Override
    public void run() {
        // 4.2. Data framing

        // Once a WebSocket connection is established, the user agent must run through the following state
        // machine for the bytes sent by the server. If at any point during these steps a read is attempted but
        // fails because the WebSocket connection is closed, then abort.

        try {
            while (active.get() && !Thread.currentThread().isInterrupted()) {
                // 4.2.1. Try to read a byte from the server. Let /frame type/ be that
                // byte.
                int frameType = inputStream.read();

                if (frameType == -1) {
                    break;
                }

                log.debug("Frame Type: {}", frameType);

                // 4.2.2. Let /error/ be false.
                boolean error = false;

                // 4.2.3. Handle the /frame type/ byte as follows:

                if ((frameType & 0x80) == 0x80) {
                    // If the high-order bit of the /frame type/ byte is set (i.e. if /frame type/ _and_ed with 0x80
                    // returns 0x80)
                    // Run these steps:

                    long length = readLength();

                    // 4.2.3.a.6. Read /length/ bytes.
                    //
                    // !!! WARNING: It is possible for a server to (innocently or maliciously) send frames with
                    // lengths greater tha 2**(31) or 2**(32) bytes, overflowing a signed or unsigned 32bit
                    // integer. User agents may therefore impose implementation-specific limits on the lengths of
                    // invalid frames that they will skip; even supporting frames 2GB in length is considered, at
                    // the time of writing, as going well above and beyond the call of duty.

                    // 4.2.3.a.7. Discard the read bytes.

                    for (long i = 0; i < length; i++) {
                        inputStream.read();
                    }

                    // 4.2.3.a.8. If the /frame type/ is 0xFF and the /length/ was 0, then run the following
                    // substeps:
                    if (frameType == 0xff && length == 0) {
                        // 4.2.3.a.8.1. If the WebSocket closing handshake has not yet started, then start
                        // the WebSocket closing handshake.
                        if (ws.startClosingHandshake()) {
                            log.debug("Closing Handshake started");
                        } else {
                            log.debug("Closing Handshake has already started");
                        }

                        // 2. Wait until either the WebSocket closing handshake has started or the WebSocket
                        // connection is closed.

                        // 3. If the WebSocket connection is not already closed, then close the WebSocket
                        // connection: *The WebSocket closing handshake has finished*. (If the connection
                        // closes before this happens, then the closing handshake doesn't finish.)

                        // 4. Abort these steps. Any data on the connection after // the 0xFF frame is
                        // discarded.
                        return;
                    } else {
                        // Otherwise, let /error/ be true.
                        log.debug("0xFF frame but lenght {} ", length);
                        error = true;
                    }
                } else if ((frameType & 0x80) == 0x0) {
                    // If the high-order bit of the /frame type/ byte is _not_ set (i.e. if /frame type/ _and_ed
                    // with 0x80 returns 0x00)
                    // Run these steps:

                    // 4.2.3.b.8.1. Let /raw data/ be an empty byte array.
                    ByteArrayOutputStream rawData = new ByteArrayOutputStream();

                    // 4.2.3.b.8.2. _Data_: Read a byte, let /b/ be that byte.
                    int b;
                    while (true) {
                        b = inputStream.read();

                        // 4.2.3.a.8.3. If /b/ is not 0xFF, then append /b/ to /raw data/ and return to the
                        // previous step (labeled _data_).
                        if (b == 0xFF) {
                            break;
                        }
                        rawData.write(b);
                    }

                    // 4.2.3.b.8.4. Interpret /raw data/ as a UTF-8 string, and store that string in /data/.
                    String data = new String(rawData.toByteArray(), UTF8_CHARSET);

                    // 4.2.3.b.8.5. If /frame type/ is 0x00, then *a WebSocket message has been received* with
                    // text /data/. Otherwise, discard the data and let /error/ be true.

                    if (frameType == 0x00) {
                        ws.onMessage(data);
                    } else {
                        error = true;
                        log.debug("Frame Type: {}",  frameType);
                    }
                }

                // 4. If /error/ is true, then *a WebSocket error has been detected*.
                if (error) {
                    throw new WebSocketException("Error detected");
                }

                // 5. Return to the first step to read the next byte.

                // If the user agent is faced with content that is too large to be handled appropriately, runs
                // out of resources for buffering incoming data, or hits an artificial resource limit intended to
                // avoid resource starvation, then it must fail the WebSocket connection.
            }
        } catch (Exception ex) {
            if (!ws.isClosed()) {
                ws.onError(ex);
                log.error(ex.getMessage(), ex);
            }
            if (ws.isReconnectEnabled()) {
                try {
                    ws.reconnect();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    private long readLength() throws IOException {
        // 4.2.3.a.1. Let /length/ be zero.
        long length = 0;

        // 4.2.3.a.2. _Length_: Read a byte, let /b/ be that byte.
        int b;
        do {
            b = inputStream.read();

            if (b == -1) {
                log.debug("End of stream occured while reading length");
                return 0;
            }

            // 4.2.3.a.3. Let /b_v/ be an integer corresponding to the low 7 bits of /b/ (the value you
            // would get by _and_ing /b/ with 0x7F).
            int b_v = b & 0x7f;

            // 4.2.3.a.4. Multiply /length/ by 128, add /b_v/ to that result, and store the final
            // result in /length/.
            length = length * 128 + b_v;

            // 4.2.3.a.5. If the high-order bit of /b/ is set (i.e. if /b/ _and_ed with 0x80 returns
            // 0x80), then return to the step above labeled _length_.
        } while ((b & 0x80) != 0x80);

        return length;
    }
}